/**
 * @file gpb.hpp
 * @date 2016-07-17
 */

// Copyright (C) 2016 Usagi
// Licensed under the MIT License.

// for Siv3D March 2016

#pragma once

#include <Siv3D.hpp>

#define BUFBYTE (4096)
#define LARGE_ABS (9999999.0)

typedef unsigned char u8;


struct GpbChunk {
	/// <summary></summary>
	String name;
	/// <summary>^Cv</summary>
	int type;
	/// <summary>t@C̎f[^ʒu</summary>
	int offsetTop;


	/// <summary>wb_[̈̕ʒu</summary>
	int offsetWrite;
};

	/// <summary>_</summary>
	struct GpbVertex {
		/// <summary>ʒu</summary>
		Vec3 position;
		/// <summary>@</summary>
		Vec3 normal;
		/// <summary>@</summary>
		Vec2 texcoord;

		/// <summary>TEXCOORD2(gp)</summary>
		Vec4 texcoord2;
	};

	/// <summary>ގ</summary>
	struct GpbMaterial {
		/// <summary>gU</summary>
		ColorF diffuse;

		/// <summary>XyL</summary>
		double power;
		/// <summary>XyLF</summary>
		ColorF specular;
		/// <summary></summary>
		ColorF ambient;

		/// <summary>255: , 0: toon01.bmp, 1: toon02.bmp, ..., 9: toon10.bmp</summary>
		int toonIndex = 255;
		/// <summary>֊sAe</summary>
		int edgeFlag = 0;

		/// <summary>ގɊ܂܂钸_</summary>
		int matVertexNum = 0;
		/// <summary>eNX`t@C</summary>
		String texName;
	};

	/// <summary>{[</summary>
	struct GpbBone {
		/// <summary></summary>
		String name;
		/// <summary>en</summary>
		String nameE;

		/// <summary>e{[̃CfbNXAꍇ-1</summary>
		int parent = -1;

		/// <summary>\Aꍇ0</summary>
		int child = 0;
		/// <summary></summary>
		int type;
		/// <summary>IKȂǂ̐eAꍇ0</summary>
		int optionParent = 0;

		/// <summary>ʒu</summary>
		Vec3 position;
	};

	/// <summary>[t</summary>
	struct GpbMorph {
		/// <summary>(20oCg܂)</summary>
		String name;
		/// <summary>en</summary>
		String nameE;

		/// <summary>0:base, 1:܂, 2:, 3:bv, 4:̑</summary>
		int type;

		/// <summary>_ԍ</summary>
		Array<int> indices;
		/// <summary>baseȂ玩gAbaseȊO͑</summary>
		Array<Vec3> vertices;
	};

	/// <summary>1̃t@C\</summary>
	struct GpbFile {
		/// <summary>o[W</summary>
		double ver = 1.0;

		/// <summary>f</summary>
		String name;
		/// <summary>gBModel name</summary>
		String nameE = L"obj";

		/// <summary>`N</summary>
		Array<GpbChunk> chunks;

		/// <summary>_</summary>
		Array<GpbVertex> vertices;

		/// <summary>ʒ_CfbNX</summary>
		Array<int> faces;

		/// <summary>eNX`</summary>
		Array<String> texNames;
		/// <summary>ގ</summary>
		Array<GpbMaterial> materials;

		/// <summary>{[</summary>
		Array<GpbBone> bones;

		/// <summary>[t</summary>
		Array<GpbMorph> morphs;

		/// <summary>zNA</summary>
		void clear() {
			this->vertices.size();
			this->faces.size();
			this->texNames.size();
			this->materials.size();
			this->bones.clear();
			this->morphs.clear();
		}

		/// <summary>{[CfbNX擾B
		/// ŏɌ̂̂݁B畉ԂB</summary>
		int getBoneIndexByName(String name) const {
			int index = -1;
			for (size_t i = 0; i < this->bones.size(); ++i) {
				if (this->bones[i].name == name) {
					index = i;
					break;
				}
			}
			return index;
		}
	};

	/// <summary>t@C</summary>
	struct GpbFiler {
		/// <summary></summary>
		/// <param name="w">oCiC^[</param>
		/// <param name="s"></param>
		void ws(BinaryWriter& w, String s) const {
			char buf[BUFBYTE];

			std::string ms = s.narrow();
			int len = ::strlen(ms.c_str());

			int size = Min(len, BUFBYTE - 4);
			::memcpy(buf, &size, 4);
			::memcpy(buf + 4, ms.c_str(), size);

			w.write(buf, (4 + size));
		}

		/// <summary>.gpbt@Co͂</summary>
		/// <param name="f">.gpb̏</param>
		/// <param name="w">oCiC^</param>
		bool SaveGpb(GpbFile& f, BinaryWriter& w) const {
			int isAnim = 0;
			int isAttr2 = 0;
			String objName = f.nameE;

			//// wb_
			w.write<u8>(0xAB);
			w.write<u8>('G');
			w.write<u8>('P');
			w.write<u8>('B');
			w.write<u8>(0xBB);
			w.write<u8>(0x0D);
			w.write<u8>(0x0A);
			w.write<u8>(0x1A);
			w.write<u8>(0x0A);

			// ver
			w.write<u8>(0x01);
			w.write<u8>(0x02);


			{
				f.chunks.clear();
				{
					GpbChunk chunk;
					chunk.name = L"scene";
					chunk.type = 1;
					chunk.offsetTop = 0;
					chunk.offsetWrite = 0;
					f.chunks.push_back(chunk);
				}
				{
					GpbChunk chunk;
					chunk.name = objName;
					chunk.type = 2;
					chunk.offsetTop = 0;
					chunk.offsetWrite = 0;
					f.chunks.push_back(chunk);
				}
				{
					GpbChunk chunk;
					chunk.name = Format(PyFmt, L"{}_Mesh", objName);
					chunk.type = 34;
					chunk.offsetTop = 0;
					chunk.offsetWrite = 0;
					f.chunks.push_back(chunk);
				}
				if (isAnim) {
					GpbChunk chunk;
					chunk.name = L"__Animations__";
					chunk.type = 3;
					chunk.offsetTop = 0;
					chunk.offsetWrite = 0;
					f.chunks.push_back(chunk);
				}
			}

			w.write<int>(f.chunks.size());
			for (auto& chunk : f.chunks) {
				ws(w, chunk.name);
				w.write<int>(chunk.type);
				chunk.offsetWrite = w.getPos();
				w.write<int>(0);
			}

			// 

			int meshNum = 1;
			w.write<int>(meshNum);
			// wb_܂


			//// bVf[^
			f.chunks[2].offsetTop = w.getPos();

			Array<Point> attrNums;
			attrNums.push_back({ 1,3 });
			attrNums.push_back({ 2,3 });
			attrNums.push_back({ 8,2 });
			if (isAttr2) {
				attrNums.push_back({ 9,4 });
			}

			w.write<int>(attrNums.size());

			int attrNumSum = 0;
			for (const auto& un : attrNums) {
				w.write<int>(un.x);
				w.write<int>(un.y);
				attrNumSum += un.y;
			}


			Vec3 mini(LARGE_ABS, LARGE_ABS, LARGE_ABS);
			Vec3 maxi(-LARGE_ABS, -LARGE_ABS, -LARGE_ABS);
			Vec3 center;

			int vNum = f.vertices.size();
			w.write<int>(sizeof(float) * attrNumSum * vNum);
			for (const auto& v : f.vertices) {
				w.write<float>(v.position.x);
				w.write<float>(v.position.y);
				w.write<float>(v.position.z);

				w.write<float>(v.normal.x);
				w.write<float>(v.normal.y);
				w.write<float>(v.normal.z);

				w.write<float>(v.texcoord.x);
				w.write<float>(v.texcoord.y);

				if (isAttr2) {
					w.write<float>(v.texcoord2.x);
					w.write<float>(v.texcoord2.y);
					w.write<float>(v.texcoord2.z);
					w.write<float>(v.texcoord2.w);
				}

				// XV
				mini.x = s3d::Min(mini.x, v.position.x);
				mini.y = s3d::Min(mini.y, v.position.y);
				mini.z = s3d::Min(mini.z, v.position.z);
				maxi.x = s3d::Max(maxi.x, v.position.x);
				maxi.y = s3d::Max(maxi.y, v.position.y);
				maxi.z = s3d::Max(maxi.z, v.position.z);
			}

// SƔaǂ
			center = mini.movedBy(maxi);
			center.setLength(center.length() * 0.5);
			double radius2 = 0.0;
			for (const auto& v : f.vertices) {
				auto rr = v.position.distanceFromSq(center);
				radius2 = s3d::Max(rr, radius2);
			}

			// ŏőȂ
			w.write<float>(mini.x);
			w.write<float>(mini.y);
			w.write<float>(mini.z);

			w.write<float>(maxi.x);
			w.write<float>(maxi.y);
			w.write<float>(maxi.z);

			w.write<float>(center.x);
			w.write<float>(center.y);
			w.write<float>(center.z);

			w.write<float>(s3d::Math::Sqrt(radius2));


			////  mesh parts
			w.write<int>(1);

			// type GL_TRIANGLES
			w.write<int>(4);
			// format GL_UNSIGNED_SHORT INDEX16
			w.write<int>(0x00001403);

			// _CfbNX
			auto faceVNum = f.faces.size();
			w.write<int>(faceVNum * sizeof(unsigned short));
			for (const auto& idx : f.faces) {
				w.write<unsigned short>(idx);
			}

			w.write<int>(1);


			//// V[
			f.chunks[0].offsetTop = w.getPos();
			// 1
			w.write<int>(1);


			//// m[ĥ悤Ȃ
			f.chunks[1].offsetTop = w.getPos();
			// 1:node, 2:joint
			w.write<int>(1);

			// transform
			for (int i = 0; i < 4; ++i) {
				for (int j = 0; j < 4; ++j) {
					w.write<float>((i == j) ? 1.0f : 0.0f);
				}
			}

			//// parent id
			ws(w, L"");
			// child
			w.write<int>(0);

			// camera
			w.write<u8>(0);
			// light
			w.write<u8>(0);

			// f
			ws(w, Format(PyFmt, L"#{}_Mesh", objName));
			// XLtO
			w.write<u8>(0);
			// ގJEg
			w.write<int>(0);


			//// Aj[V
			if (isAnim) {
				f.chunks[3].offsetTop = w.getPos();

				w.write<int>(1);

				ws(w, L"animations");

				w.write<int>(1);

				{
					ws(w, objName);
					w.write<int>(0x10);

					int keyNum = 30 + 1;

					w.write<int>(keyNum);
					for (int i = 0; i < keyNum; ++i) {
						w.write<int>(i * 1000 / 30);
					}

					w.write<int>(keyNum * 7);
					for (int i = 0; i < keyNum; ++i) {
						w.write<float>(0.0f);
						w.write<float>(0.0f);
						w.write<float>(0.0f);
						w.write<float>(1.0f);

						w.write<float>(0.0f);
						w.write<float>(0.0f);
						w.write<float>(0.0f);
					}

					w.write<int>(0);

					w.write<int>(0);

					w.write<int>(1);
					w.write<int>(1);
				}
			}

			{
				ws(w, L"");
				w.write<float>(0.0f);
				w.write<float>(0.0f);
				w.write<float>(0.0f);
			}


			//// ߂ď
			for (const auto& chunk : f.chunks) {
				w.setPos(chunk.offsetWrite);
				w.write<int>(chunk.offsetTop);
			}

			w.close();
			return true;
		}

		/// <summary>}eAt@Co</summary>
		bool SaveMaterial(GpbFile& f, TextWriter w) const {
			String templateFileName = L"template.material";
			TextReader tr(templateFileName, s3d::TextEncoding::UTF8);

			String whole;
			auto result = tr.readAll(whole);
			if (result == false) {
				return false;
			}

			for (const auto& m : f.materials) {
				String s = whole.replace(L"__NAME00__", f.nameE);
				s = s.replace(L"__TEXFILE00__", s3d::FileSystem::FileName(m.texName));
				s = s.replace(L"__POWER00__", Format(m.power));

				w.writeln(s);

				break;
			}

			w.close();
			return true;
		}


		/// <summary></summary>
		bool LoadGpb(BinaryReader& r, GpbFile& f) const {
			f.chunks.clear();

			int v32;

			// wb_
			r.skip(11);

			int cNum = 0;
			r.read<int>(cNum);

			for (int i = 0; i < cNum; ++i) {
				GpbChunk chunk;

				chunk.name = rs(r);
				r.read<int>(chunk.type);
				r.read<int>(chunk.offsetTop);

				f.chunks.push_back(chunk);

				Println(PyFmt, L"{},{},{}",
					chunk.name, chunk.type, chunk.offsetTop);
			}

			auto cur = r.getPos();
			r.read<int>(v32);
			Println(PyFmt, L"listTail,{}, s32,{}", cur, v32);

			for (int i = 0; i < cNum; ++i) {
				r.setPos(f.chunks[i].offsetTop);
				
				switch (f.chunks[i].type) {
				case 1: // V[
					r.read<int>(v32);
					Println(PyFmt, L"V[,{}", v32);
					break;
				case 2:
					break;
				case 3: // Aj[V
					break;
				case 32: // J
					break;
				case 33: // Cg
					break;
				case 34: // bV
					LoadMesh(r, f);
					break;
				case 128: // tHg
					break;
				}

			}

			return true;
		};

		/// <summary>bV̓ǂݍ݁B</summary>
		bool LoadMesh(BinaryReader& r, GpbFile& f) const {
			int v32 = 0;
			float vf = 0.0f;

			int num = 0;
			r.read<int>(num);

			int attrNumSum = 0;
			for(int i = 0; i < num; ++i) {
				r.read<int>(v32);
				r.read<int>(v32);
				attrNumSum += v32;
			}

			r.read<int>(num);
			int vNum = num / sizeof(float) / attrNumSum;
			for (int i = 0; i < vNum; ++i) {
				for (int j = 0; j < attrNumSum; ++j) {
					r.read<float>(vf);
				}
			}

			Vec3 mini = readF3(r);
			Vec3 maxi = readF3(r);
			Vec3 center = readF3(r);
			float radius;
			r.read<float>(radius);

			Println(mini);
			Println(maxi);
			Println(center);
			Println(PyFmt, L"radius,{}", radius);

			return true;
		};

		/// <summary>^ǂݎ</summary>
		s3d::String rs(BinaryReader& r) const {
			char buf[BUFBYTE];
			int v32 = 0;
			r.read<int>(v32);

			v32 = s3d::Clamp<int>(v32, 0, BUFBYTE - 1);

			r.read(buf, v32);
			buf[v32] = 0;
			std::string s0(buf);
			return Widen(s0);
		};

		/// <summary>float3ǂݎ</summary>
		s3d::Vec3 readF3(BinaryReader& r) const {
			s3d::Vec3 ret;
			float vf;
			r.read<float>(vf);
			ret.x = vf;
			r.read<float>(vf);
			ret.y = vf;
			r.read<float>(vf);
			ret.z = vf;
			return ret;
		};
	};
